#!/usr/bin/env python2

#    Copyright (C) 2011  Pablo Fernandez Fernandez <pfernandez 13t org>
#
#    GridBob is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GridBob is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GridBob.  If not, see <http://www.gnu.org/licenses/>.

# Interface module to Cream
# Provides:
#  - IsReady ()
#  - SubmitJob (job) -> jobid
#  - QueryJob (jobid) 
#  - CancelJob (jobid)
#  - GetResult (jobid) -> job

from subprocess import STDOUT, PIPE, Popen
import os, sys, tempfile, re

# Job states (for our application):
# QUEUED, RUNNING, DONE, FAILED

# Job states (for Cream):
# REGISTERED, PENDING, IDLE, HELD, RUNNING, REALLY-RUNNING  (temporary states)
# DONE-OK, CANCELLED, DONE-FAILED, ABORTED, UNKNOWN  (terminal states)


def error (text):
  sys.stderr.write(text+"\n")
  sys.exit(1)

class CE:
  def __init__ (self, url):
    # Normal shape: cream.domain.org:8443/cream-pbs-test
    self.url = url
    self.enablesubmit = False
    # We accept black CE, but only for querying jobs.
    if url == "":  
      self.onlytoqueryjobs = True
    else:
      fields = url.split("/")
      if len (fields) != 2 or len (fields[0].split(".")) < 3:
        error ("CE URL needs to have a format like: cream.domain.org:8443/cream-pbs-test")
      self.host_port = fields[0] # we don't care if they put the port or not (theoretically)
      self.onlytoqueryjobs = False

  def isready (self):
    if self.onlytoqueryjobs:
      error ("Tried to call a CE function that requires proper CE initialization")
    proc = Popen(["glite-ce-allowed-submission", self.host_port ], stdout=PIPE, stderr=STDOUT)    
    output = proc.communicate()[0]
    if output.split()[-1] == 'enabled':
      return True
    else:
      print output
      return False

  def submit(self, min, max, ceurl, session):
    if self.onlytoqueryjobs:
      error ("Tried to call a CE function that requires proper CE initialization")
    tmpfile = tempfile.NamedTemporaryFile ('w')
    tmpfile.write('[ Type = "job"; JobType = "normal"; Executable = "executor.sh"; Arguments = "%d %d"; StdOutput = "stdout"; StdError = "stdout"; OutputSandbox = {"stdout", "john.pot"}; OutputSandboxBaseDestURI = "gsiftp://localhost"; InputSandbox = {"lib/executor.sh", "%s", "%s", "%s"}; ]' % (min, max, session.passwdfile, session.runtimefile, session.resultsfile))
    tmpfile.flush()
    proc = Popen(["glite-ce-job-submit", "-a", "-r", ceurl, tmpfile.name ], stdout=PIPE, stderr=STDOUT)
    output = proc.communicate()[0]
    tmpfile.close()  # deletes the file automatically
    if proc.returncode == 0:
      return (True, output.splitlines()[0])
    else:
      return (False, output)


  def checkoutput (self, jobid):
    proc = Popen(["glite-ce-job-status", "-L2", jobid], stdout=PIPE, stderr=STDOUT)
    output = proc.communicate()[0]
    if proc.returncode != 0:
      error ("Failed to query job, the jobid could be wrong: %s" % jobid)
    m = re.search ('Current Status = \[(.*)\]', output)
    if not m:
      error ("The output is not correctly formatted:\n %s" % output)
    status = m.group(1)
    if status == 'DONE-OK':
      # Additionally, we check the exit code. if != 0, there was a problem, treat as FAILED!
      m = re.search ('ExitCode       = \[(.*)\]', output)
      if m:
        exitcode = int(m.group(1))
        if exitcode > 1:
          return {'state': 'FAILED', 'reason': "Exit value returned not ok: %d" % exitcode}
      # The job has finished, Let's take a look at the Output Sandbox (OSB)
      m = re.search ('CREAM OSB URI  = \[(.*)\]', output)
      if not m:
        error ("The output does not have CREAM OSB URI:\n %s" % output)
      osb = m.group(1) 
      # Connect to gridftp to get the output passwords
      proc = Popen(["globus-url-copy", osb + '/john.pot', "-"], stdout=PIPE, stderr=STDOUT)
      output = proc.communicate()[0]
      if proc.returncode != 0:
        error ("Error retrieving: %s" % osb + '/john.pot')
      else:
        value = output.splitlines()
        # Remove empty lines
        for i in range(0, value.count('')):
          value.remove('')
        # We have the passwords, let's try to get the time spent
        proc = Popen(["globus-url-copy", osb + '/stdout', "-"], stdout=PIPE, stderr=STDOUT)
        stdout = proc.communicate()[0]
        if proc.returncode != 0:
          runtime = 0
        else:
          m = re.search ('\[RUNTIME=([0-9]+)\]', stdout)
          if m:
            runtime = int(m.group(1))
          else:
            runtime = 0
        # Finally return with all the information
        return {'state': 'DONE', 'passwords': value, 'runtime': runtime}
    elif status == 'DONE-FAILED' or status == 'CANCELLED' or status == 'ABORTED':
      m = re.search ('FailureReason  = \[(.*)\]', output)
      if m:
        return {'state': 'FAILED', 'reason': m.group(1)}
      else:
        return {'state': 'FAILED', 'reason': 'Could not parse the failure reason. %s' % status}
    elif status == 'RUNNING' or status == 'REALLY-RUNNING':
      return {'state': 'RUNNING'}
    elif status == 'REGISTERED' or status == 'PENDING' or status == 'IDLE' or status == 'HELD':
      return {'state': 'QUEUED'}
    else:
      # The job was either failed or not done
      print output
      return (False, {'state': 'unknown'})



  def showinfo (self, jobid):
    print "Parsing: glite-ce-job-status -L2 %s" % jobid
    proc = Popen(["glite-ce-job-status", "-L2", jobid], stdout=PIPE, stderr=STDOUT)
    output = proc.communicate()[0]
    if proc.returncode != 0:
      error ("Failed to query job, the jobid could be wrong: %s" % jobid)
    m = re.search ('Current Status = \[(.*)\]', output)
    if not m:
      error ("The output is not correctly formatted:\n %s" % output)
    status = m.group(1)

    if status == 'DONE-OK':
      # Additionally, we check the exit code. if != 0, there was a problem, treat as FAILED!
      m = re.search ('ExitCode       = \[(.*)\]', output)
      if not m:
        error ("Can't find ExitCode:\n %s" % output)
      exitcode= m.group(1)

    # Check if its in terminal status:
    if status == 'DONE-OK' or status == 'DONE-FAILED' or status == 'CANCELLED' or status == 'ABORTED':
      print "***** JOB HAS FINISHED: %s *****" % status
      if status == 'DONE-OK':
        # Additionally, we check the exit code. if != 0, there was a problem, treat as FAILED!
        m = re.search ('ExitCode       = \[(.*)\]', output)
        if not m:
          error ("Can't find ExitCode:\n %s" % output)
        exitcode = int(m.group(1))
        print "***** EXIT VALUE: %d *****" % exitcode
      # The job has finished, Let's take a look at the Output Sandbox (OSB)
      m = re.search ('CREAM OSB URI  = \[(.*)\]', output)
      if not m:
        error ("The output does not have CREAM OSB URI:\n %s" % output)
      osb = m.group(1)
      # Connect to gridftp to get the standard output
      print "***** JOB STDOUT *****"
      proc = Popen(["globus-url-copy", osb + '/stdout', "-"], stdout=PIPE, stderr=STDOUT)
      stdout = proc.communicate()[0]
      if proc.returncode != 0:
        error ("Error retrieving: %s" % osb + '/stdout')
      else:
        print stdout
      # Check if there is a Failure Reason
      m = re.search ('FailureReason  = \[(.*)\]', output)
      if m:
        print "***** CREAM FAILURE REASON *****"
        print m.group(1)
    elif status == 'RUNNING' or status == 'REALLY-RUNNING' or status == 'REGISTERED' or status == 'PENDING' or status == 'IDLE' or status == 'HELD':
      print "***** JOB DID NOT FINISH YET: %s *****" % status
    else:
      # The job was either failed or not done
      print "***** JOB IN AN UNKNOWN STATE: %s *****" % status
      print output
    # We print the worker node that it ran in
    m = re.search ('Worker Node.*= \[(.*)\]', output)
    if m:
      print "  Worker Node: %s" % m.group(1)
    # In any way Now we print status transition
    #m = re.search ('(Job status changes.*)Issued Commands', output)
    m = re.search ('(Job status changes.*)Issued Comm', output, flags=re.MULTILINE|re.DOTALL)
    if m:
      print "  "+m.group(1)
