#!/usr/bin/env python2

#    Copyright (C) 2011  Pablo Fernandez Fernandez <pfernandez 13t org>
#
#    GridBob is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GridBob is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GridBob.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
from subprocess import STDOUT, PIPE, Popen
import datetime

try:
    import sqlite3
except:
    from pysqlite2 import dbapi2 as sqlite3

# Job states:
# QUEUED, RUNNING, DONE, FAILED


# Quick error function
def error (text):
  sys.stderr.write(text+"\n")
  sys.exit(1)

class Session:
  def __init__ (self, name):
    self.basepath = os.path.abspath(os.path.dirname(__file__) + "/../")
    self.latestsessionfile = self.basepath + "/session/latest"
    # We check if a name is given, otherwise we use "latest" symlink
    if name == "latest": error ("Invalid session name")
    elif name == "":
      try:
        fd = open(self.latestsessionfile, 'r')
        self.name = fd.read().splitlines()[0]
        fd.close()
      except: 
        error ("Unknown session name, and no default set.")
    else: 
      self.name = name
    # Files/dirs
    self.directory = self.basepath + "/session/" + self.name
    self.defaultblocksizefile = self.directory + "/default_blocksize"
    self.defaultceurlfile = self.directory + "/default_ceurl"
    self.runtimefile = self.directory + "/runtime.tgz"
    self.dbfile = self.directory + "/jobs.sqlite"
    self.resultsfile = self.directory + "/results"
    self.passwdfile = self.directory + "/passwd"
    # Values
    self.defaultblocksize = False
    self.defaultceurl = False
    self.hashedpasswords = []
    self.discoveredpasswords = []
    self.jobs = []
    self.loaded = False  # Set to true when we load an existing session
    self.jobstates = ('QUEUED', 'RUNNING', 'DONE', 'FAILED')

  # Remove the session and everything inside
  def destroy(self):
    for file in os.listdir(self.directory):
      os.remove(self.directory +"/"+ file)
    os.rmdir(self.directory)

  def __readfile(self, file):
    if not os.access (file, os.R_OK):
      error ("Session file not readable: %s" % file)
    try:
      fd = open(file, 'r')
      value = fd.read().splitlines()
      # Remove empty lines
      for i in range(0, value.count('')):
        value.remove('')
      fd.close()
    except:
      print "Unexpected error: reading %s\n " % file
      raise
    return value

  def __writefile(self, file, content):
    try:
      fd = open(file, 'w')
      for i in content:
        fd.write(str(i) + "\n")
      fd.close()
    except:
      print "Unexpected error: writing %s\n " % file
      raise

  # Read values inside the session
  def load(self):
    # Define the latest opened session as the current one.
    self.__writefile(self.latestsessionfile, [self.name])
    # Read needed mandatory session files
    for file in (self.runtimefile, self.dbfile):
      if not os.access (file, os.R_OK):
        error ("Session file not readable: %s" % file)
    self.hashedpasswords = self.__readfile(self.passwdfile)
    self.discoveredpasswords = self.__readfile(self.resultsfile)
    # Read optional session files
    if os.access (self.defaultblocksizefile, os.R_OK):
      self.defaultblocksize = int(self.__readfile(self.defaultblocksizefile)[0])
    if os.access (self.defaultceurlfile, os.R_OK):
      self.defaultceurl = self.__readfile(self.defaultceurlfile)[0]
    # Read job database
    self.jobs = []
    try:
      con = sqlite3.connect(self.dbfile)
      con.row_factory = sqlite3.Row
      cur = con.cursor()
      cur.execute("SELECT * FROM jobs ORDER BY num")
      for row in cur:
        self.jobs.append({'num':row['num'], 'min':row['min'], 'max':row['max'], 'jobid':row['jobid'], 'state':row['state'], 'discovered':row['discovered'], 'createtime':row['createtime'], 'runtime':row['runtime'] })
        # Check we have loaded them in order, and starting from 0
        if len(self.jobs) -1 != row["num"]:
          error ("Saved job list is not complete. Can't load.")
      con.close()
    except:
      print "Error reading jobs from database:"
      print self.jobs
      raise

    self.loaded = True

  # Create the session
  def save(self):
    # Define the latest opened session as the current one.
    self.__writefile(self.latestsessionfile, [self.name])      
    # We see if the session exists
    if not os.access (self.directory, os.W_OK):
      error ("Session does not exist, call create() instead.")
    if not self.loaded:
      # We force a self.loaded = True (means, force to be loaded first)
      error ("Tried to save a non-loaded session. Run load() before save()!")
    # Write optional stuff:
    if self.defaultblocksize:
      self.__writefile(self.defaultblocksizefile, [self.defaultblocksize])      
    if self.defaultceurl:
      self.__writefile(self.defaultceurlfile, [self.defaultceurl])
    # Mandatory files (some empty):
    self.__writefile(self.passwdfile, self.hashedpasswords)
    self.__writefile(self.resultsfile, self.discoveredpasswords)
    # Database
    try:
      con = sqlite3.connect(self.dbfile)
      con.row_factory = sqlite3.Row
      for job in self.jobs:
        con.execute("INSERT blabla INTO job")
      con.close()
    except:
      print "Unexpected error: creating sqLite database in %s\n " % dbfile
      raise
 
  def create(self):
    if os.access (self.directory, os.W_OK):
      error ("Session already exists, call save() instead")
    if not self.loaded:
      # We force a self.loaded = True
      error ("Tried to save a session with no contents. Do 'self.loaded = True' before")
    # Create the dir
    try:
      os.mkdir (self.directory)
      newdatabase = True
    except:
      print "Unexpected error: creating dir %s\n " % self.directory
      raise
    # Create the DB table:
    try:
      con = sqlite3.connect(self.dbfile)
      con.row_factory = sqlite3.Row
      con.execute("CREATE TABLE jobs (num INTEGER PRIMARY KEY, min INT, max INT, createtime TIMESTAMP, state TEXT, discovered INT, jobid TEXT, runtime INT)")
      con.commit()
      con.close()
    except:
      print "Unexpected error: creating sqLite database in %s\n " % dbfile
      raise
    # Now we write the data
    self.save()

    
  def status(self):
    # Write a summary
    print "SESSION NAME: %s" % self.name
    print "  Directory (no need to touch there): %s" % self.directory
    print "  Hashed passwords loaded: %d" % len(self.hashedpasswords)
    print "  Discovered passwords: %d" % len(self.discoveredpasswords)
    if self.defaultblocksize:
      print "  Default block size: %s" % self.defaultblocksize
    if self.defaultceurl:
      print "  Default CE url: %s" % self.defaultceurl
    print "  Current defined jobs: %d" % len(self.jobs)
    # Count the number of jobs per state:
    # First set all to 0
    number = {}
    for s in self.jobstates:
      number[s] = 0
    # Then count
    for job in self.jobs:
      number[job["state"]] += 1
    # And then print only >0 values
    for state in self.jobstates:
      if number[state] > 0:
        print "    %s: %d jobs" % (state, number[state])


  def printjobs(self):
    for job in self.jobs:
      if job["jobid"] != "Fake job":
        print "%d - %s - %s - %.1f min" % (job["num"], job["jobid"], job["state"], job["runtime"]/60.0)


  # passwd file (called upon session creation only)
  def loadpasswdfromfile(self, passwdfile):
    output = self.__readfile(passwdfile)
    for line in output:
      fields = line.split(":")
      if len(fields) > 1 and len(fields[1]) > 6:  #We consider it a hashed password, to avoid empty passwords
        self.hashedpasswords.append(fields[0] + ":" + fields[1])
    
  # Runtime John directory and binary. Must be run after save()
  def loadruntime(self, runtimedir):
    if not os.access (self.directory, os.W_OK):
      error ("Cannot load runtime before the first save() of the session")
    if os.access (runtimedir + "/john", os.X_OK):
      proc = Popen(["tar", "-czf", self.runtimefile, "-C", runtimedir, "."], stdout=PIPE, stderr=STDOUT)    
      output = proc.communicate()[0]
      if proc.returncode != 0:
        error ("Unexpected error: failed to run tar -czf %s -C %s ." % (self.runtimefile, runtimedir))
    else:
      error ("Error accesing DIR %s. John's executable not found!" % options.runtime)

  def newjob(self, num, min, max, jobid):
    now = datetime.datetime.now().isoformat(' ')
    self.jobs.append({'num':num, 'min':min, 'max':max, 'jobid':jobid, 'state':'QUEUED', 'discovered':0, 'createtime':now })
    try:
      con = sqlite3.connect(self.dbfile)
      con.row_factory = sqlite3.Row
      con.execute("INSERT INTO jobs (num, min, max, createtime, state, discovered, jobid, runtime) VALUES (?,?,?,?,?,0,?,0)", (num, min, max, now, 'QUEUED', jobid) )
      con.commit()
      con.close()
    except:
      print "Error inserting job into database:"
      print self.jobs[num]
      raise

  def updatejob(self, num, state, result, jobid, runtime):
    self.jobs[num]["state"] = state
    self.jobs[num]["discovered"] = len(result)
    self.jobs[num]["jobid"] = jobid
    self.jobs[num]["runtime"] = runtime
    if len(result) > 0:
      for password in result:
        if password not in self.discoveredpasswords:
          self.discoveredpasswords.append(password)
      self.__writefile(self.resultsfile, self.discoveredpasswords)
    try:
      con = sqlite3.connect(self.dbfile)
      con.row_factory = sqlite3.Row
      con.execute("UPDATE jobs SET state=?, discovered=?, jobid=?, runtime=? WHERE num=?", (state, len(result), jobid, runtime, num) )
      con.commit()
      if con.total_changes == 0:
        print "Error: no database register was modified:"
      con.close()
    except:
      print "Error updating job into database:"
      print self.jobs[num]
      raise
 


